/*
 *  GOL - Game of life simulation software
 *  Copyright (C) 1996 Michael Raymond Brown
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Michael Raymond Brown
 *  EMail:  michael@arrakeen.demon.co.uk
 *  WWW:    http://www.arrakeen.demon.co.uk/
 *  Address:  Michael Brown
 *            Kia Ora
 *            Brook Street
 *            Great Bromley
 *            Nr Colchester
 *            Essex
 *            CO7 7HX
 *            England
 */


#include "png_life.h"
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <new.h>

extern "C" {
#include <png.h>
}

extern const char* lasterrstr;
extern int	   lasterrnum;

static char keywords[] = "Game of life rules";

/**** "Game of life rules text segment"
 *  1 byte of 'y' or 'n' for wrap arround at edges
 *  9 bytes of 'x' or ' ' for the mask
 * 10 bytes of 'c'/'s'/l' for the rules (clear,set,leave)
 *    ie conways = y-xxxx xxxx-cclscccccc
 */


static unsigned char reverse[256]=
	{   0, 128,  64, 192,  32, 160,  96, 224
	,  16, 144,  80, 208,  48, 176, 112, 240
	,   8, 136,  72, 200,  40, 168, 104, 232
	,  24, 152,  88, 216,  56, 184, 120, 248
	,   4, 132,  68, 196,  36, 164, 100, 228
	,  20, 148,  84, 212,  52, 180, 116, 244
	,  12, 140,  76, 204,  44, 172, 108, 236
	,  28, 156,  92, 220,  60, 188, 124, 252
	,   2, 130,  66, 194,  34, 162,  98, 226
	,  18, 146,  82, 210,  50, 178, 114, 242
	,  10, 138,  74, 202,  42, 170, 106, 234
	,  26, 154,  90, 218,  58, 186, 122, 250
	,   6, 134,  70, 198,  38, 166, 102, 230
	,  22, 150,  86, 214,  54, 182, 118, 246
	,  14, 142,  78, 206,  46, 174, 110, 238
	,  30, 158,  94, 222,  62, 190, 126, 254
	,   1, 129,  65, 193,  33, 161,  97, 225
	,  17, 145,  81, 209,  49, 177, 113, 241
	,   9, 137,  73, 201,  41, 169, 105, 233
	,  25, 153,  89, 217,  57, 185, 121, 249
	,   5, 133,  69, 197,  37, 165, 101, 229
	,  21, 149,  85, 213,  53, 181, 117, 245
	,  13, 141,  77, 205,  45, 173, 109, 237
	,  29, 157,  93, 221,  61, 189, 125, 253
	,   3, 131,  67, 195,  35, 163,  99, 227
	,  19, 147,  83, 211,  51, 179, 115, 243
	,  11, 139,  75, 203,  43, 171, 107, 235
	,  27, 155,  91, 219,  59, 187, 123, 251
	,   7, 135,  71, 199,  39, 167, 103, 231
	,  23, 151,  87, 215,  55, 183, 119, 247
	,  15, 143,  79, 207,  47, 175, 111, 239
	,  31, 159,  95, 223,  63, 191, 127, 255};



static void get_rules(png_infop info_ptr, char* rules)
   {int count = info_ptr->num_text;
    while (count--) if (strcasecmp(info_ptr->text[count].key, keywords)==0
		     && info_ptr->text[count].text_length>=22)
       {memcpy(rules, info_ptr->text[count].text, 22);
	if (  rules[ 1]!='-'
	  || (rules[ 0]!='y' && rules[ 0]!='n')
	  || (rules[ 2]!='x' && rules[ 2]!=' ')
	  || (rules[ 3]!='x' && rules[ 3]!=' ')
	  || (rules[ 4]!='x' && rules[ 4]!=' ')
	  || (rules[ 5]!='x' && rules[ 5]!=' ')
	  || (rules[ 6]!='x' && rules[ 6]!=' ')
	  || (rules[ 7]!='x' && rules[ 7]!=' ')
	  || (rules[ 8]!='x' && rules[ 8]!=' ')
	  || (rules[ 9]!='x' && rules[ 9]!=' ')
	  || (rules[10]!='x' && rules[10]!=' ')
	   || rules[11]!='-'
	  || (rules[12]!='c' && rules[12]!='l' && rules[12]!='s')
	  || (rules[13]!='c' && rules[13]!='l' && rules[13]!='s')
	  || (rules[14]!='c' && rules[14]!='l' && rules[14]!='s')
	  || (rules[15]!='c' && rules[15]!='l' && rules[15]!='s')
	  || (rules[16]!='c' && rules[16]!='l' && rules[16]!='s')
	  || (rules[17]!='c' && rules[17]!='l' && rules[17]!='s')
	  || (rules[18]!='c' && rules[18]!='l' && rules[18]!='s')
	  || (rules[19]!='c' && rules[19]!='l' && rules[19]!='s')
	  || (rules[20]!='c' && rules[20]!='l' && rules[20]!='s')
	  || (rules[21]!='c' && rules[21]!='l' && rules[21]!='s')) continue;
	break;
       }
   }

static char staticmsg[128];
static void png_life_error(png_structp png_ptr, char* message)
   {strcpy(staticmsg, message);
    lasterrstr=staticmsg;
    lasterrnum=0;
    longjmp(png_ptr->jmpbuf, 1);
   }


   // If lifep==null, just obtain the file dimensions and life rules
   // If on return *rules='\0', there where no rules in the file
int open_png_life(Life* life, const char* fname, int* widthp, int* heightp, char* rules)
   {char* ptr = (char*)0;
    *rules='\0';

    FILE* fp = fopen(fname, "rb");
    if (!fp)
       {
	return -1;
       }

    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING
		, (void*)0, (png_error_ptr)png_life_error, (png_error_ptr)0);
    if (!png_ptr)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	fclose(fp);
	return -1;
       }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	fclose(fp);
	png_destroy_read_struct(&png_ptr, (png_infopp)0, (png_infopp)0);
	return -1;
       }

    png_infop end_info = png_create_info_struct(png_ptr);
    if (!end_info)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	fclose(fp);
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)0);
	return -1;
       }

    lasterrnum = 0;
    lasterrstr = "unknown";
    if (setjmp(png_ptr->jmpbuf))
       {png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	if (ptr) free(ptr);
	fclose(fp);
	return -1;
       }

    png_init_io(png_ptr, fp);
    png_read_info(png_ptr, info_ptr);
    *widthp  = info_ptr->width ;
    *heightp = info_ptr->height;

    if (info_ptr->bit_depth != 1)
       {lasterrnum=0;
	lasterrstr="invalid bit depth";
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	fclose(fp);
	return -1;
       }

    if (info_ptr->channels != 1)
       {lasterrnum=0;
	lasterrstr="too many color channels";
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	fclose(fp);
	return -1;
       }

    get_rules(info_ptr, rules);

    if (!life)
       {png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	fclose(fp);
	return 0;
       }

    int passes = 1;
    if (info_ptr->interlace_type)
	png_set_interlace_handling(png_ptr);

    png_read_update_info(png_ptr, info_ptr);

    if (info_ptr->rowbytes & 1) info_ptr->rowbytes++;
    png_bytep rows[16];
    ptr = (char*)malloc(info_ptr->rowbytes * 16);
    if (!ptr)
       {lasterrnum=errno;
	lasterrstr=strerror(errno);
	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
	fclose(fp);
	return -1;
       }

    int count;
    for (count=0; count<16; count++)
	{
	 rows[count] = png_bytep(ptr + (count*info_ptr->rowbytes));
	 rows[count][info_ptr->rowbytes-1]='\0';
	}


    // Do first pass here
    u_int data[16];
    int  width = (info_ptr->width +15)/16 * 2;
    int height = (info_ptr->height+15)/16 * 2;
    int   x, y;
    int goagain = 0;

    do {
	for (y=0; y<height; y+=2)
	    {
	     if (goagain)
		{for (x=0; x<width; x+=2)
		     {life->getArea(x/2, y/2, data);
		      for (count=0; count<16; count++)
			  {rows[count][x+0]=reverse[(data[count]   ) & 0xff];
			   rows[count][x+1]=reverse[(data[count]>>8) & 0xff];
			  }
		     }
		}

	     png_read_rows(png_ptr, rows, NULL, 16);
	     for (count=info_ptr->height - (y*8); count<16; count++)
		  memset(rows[count], 0, info_ptr->rowbytes);

	     for (x=0; x<width; x+=2)
		 {for (count=0; count<16; count++)
		      {data[count] =  u_int(reverse[rows[count][x+0]])
				   | (u_int(reverse[rows[count][x+1]])<<8);
		      }
		  life->setArea(x/2, y/2, data);
		 }
	    }
	goagain=1;
    } while (--passes);

    png_read_end(png_ptr, end_info);
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(fp);
    free(ptr);

    return 0;
   }




int save_png_life(Life* life, const char* fname)
   {png_bytep rows[16];
    int  width = life->getWidth()/8;
    int  count;

    char* ptr = (char*)malloc(width * 16);
    if (!ptr)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	return -1;
       }
    for (count=0; count<16; count++)
	 rows[count]=png_bytep(ptr+(count*width));

    FILE* fp = fopen(fname, "wb");
    if (!fp)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	free(ptr);
	return -1;
       }

    png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
		(void*)0, (png_error_ptr)png_life_error, (png_error_ptr)0);
    if (!png_ptr)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	fclose(fp);
	free(ptr);
	return -1;
       }

    png_infop info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
       {lasterrnum = errno;
	lasterrstr = strerror(errno);
	fclose(fp);
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	free(ptr);
	return -1;
       }

    if (setjmp(png_ptr->jmpbuf))
       {lasterrnum = 0;
	lasterrstr = "unknown";
	fclose(fp);
	png_destroy_write_struct(&png_ptr, &info_ptr);
	free(ptr);
	return -1;
       }

    png_init_io(png_ptr, fp);

    info_ptr->width  = life->getWidth();
    info_ptr->height = life->getHeight();
    info_ptr->bit_depth = 1;
    info_ptr->color_type= PNG_COLOR_TYPE_GRAY;
    info_ptr->interlace_type = 0;
    info_ptr->valid |= 0;	// I know!

    png_text text[2];
    char softtxt[128];
    char ruletxt[128];
    extern char* GOL_VERSION;
    sprintf(softtxt, "GOL (Game of life) ver %s  http://www.arrakeen.demon.co.uk/life/", GOL_VERSION);
    sprintf(ruletxt, "%s", life->rules);

    text[0].compression=-1;
    text[0].key = "Software";
    text[0].text = softtxt;
    text[0].text_length = strlen(softtxt);

    text[1].compression=-1;
    text[1].key = keywords;
    text[1].text = ruletxt;
    text[1].text_length = strlen(ruletxt);

    info_ptr->num_text = 2;
    info_ptr->text = text;

    png_write_info(png_ptr, info_ptr);

    int height;
    int   x, y;
    u_int data[16];
    height = life->getHeight()/8;
    width = life->getWidth()/8;

    for (y=0; y<height; y+=2)
	{
	 for (x=0; x<width; x+=2)
	     {life->getArea(x/2, y/2, data);
	      for (count=0; count<16; count++)
		  {rows[count][x+0]=reverse[(data[count]   ) & 0xff];
		   rows[count][x+1]=reverse[(data[count]>>8) & 0xff];
		  }
	     }
	 png_write_rows(png_ptr, rows, 16);
	}

    info_ptr->num_text=0;
    png_write_end(png_ptr, info_ptr);
    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fp);
    free(ptr);

    return 0;
   }

